var rewire = require('rewire');
var mocPlugin = rewire('../../src/plugins/mobile-optimised-content-plugin');

var globals = {
    window: {
        document: {},
        Image: 0
    }
};
mocPlugin.__set__(globals);

var widgetInstance = {
    features: {
        cxp: {
            config: {}
        }
    },
    body: {}
};

var widgetModel = {
    id: 0,
    preferences: {
        widgetContentsUpdated: {
            name: 'widgetContentsUpdated',
            value: '12345'
        }
    }
};

/*
 *  Promise based function tests
 */
describe('mobile optimised content plugin - getDataUri', function () {
    it('fetches the image - fail', function(done) {
        var mockImage = {
            onload: 0,
            onerror: 0
        };
        var Image = jasmine.createSpy('Image').and.returnValue(mockImage);
        mocPlugin.__set__('window.Image', Image);

        var getDataUri = mocPlugin.__get__('getDataUri');
        var result = getDataUri('/image.jpg');
        mockImage.onerror();

        result.catch(function() {
            done();
        })
    });

    it('fetches the image - ok', function(done) {
        var mockImage = {
            onload: 0,
            onerror: 0
        };
        var Image = jasmine.createSpy('Image').and.returnValue(mockImage);
        mocPlugin.__set__('window.Image', Image);

        var mockCanvas = {};
        var mockContext = {};
        mockContext.drawImage = jasmine.createSpy('drawImage');
        mockCanvas.getContext = jasmine.createSpy('getContext').and.returnValue(mockContext);
        mockCanvas.toDataURL = jasmine.createSpy('toDataURL').and.returnValue('base64:image');
        globals.window.document.createElement = jasmine.createSpy('createElement').and.returnValue(mockCanvas);

        var getDataUri = mocPlugin.__get__('getDataUri');
        var result = getDataUri('/image.jpg');
        mockImage.onload();

        result.then(function(data) {
            expect(data).toEqual('base64:image');
            done();
        })
    });

    // I can fake any function call here, but I don't really know how canvas reacts to certain things. So I
    // cannot test what happens with half-broken images or encoding errors.
});

describe('mobile optimised content plugin - fetchContent', function () {
    it('fetches the content - fail', function(done) {
        var fetchPromise = new Promise(function(resolve, reject) {
            reject('broken');
        });
        var fetch = jasmine.createSpy('fetch').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetch', fetch);

        var fetchContent = mocPlugin.__get__('fetchContent');
        fetchContent('/foo', 'bar', widgetModel.id)
            .catch(function() {
                done();
            });
    });

    it('fetches the content - 404', function(done) {
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve({
                status: '404'
            });
        });
        var fetch = jasmine.createSpy('fetch').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetch', fetch);

        var fetchContent = mocPlugin.__get__('fetchContent');
        fetchContent('/foo', 'bar', widgetModel.id)
            .catch(function() {
                done();
            });
    });

    it('fetches the content - ok', function(done) {
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve({
                status: '200',
                text: function() {
                    return 'content template'
                }
            });
        });
        var fetch = jasmine.createSpy('fetch').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetch', fetch);

        var fetchContent = mocPlugin.__get__('fetchContent');
        fetchContent('/foo', 'bar', widgetModel.id)
            .then(function(content) {
                expect(content).toEqual('content template');
                done();
            });
    });

    it('finds CS references (images) and tries to resolve them - fails', function(done) {
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve({
                status: '200',
                text: function() {
                    return 'content template - this is an image reference: "/foo/content/bbp/repositories/image.jpg"'
                }
            });
        });
        var fetch = jasmine.createSpy('fetch').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetch', fetch);

        var imagePromise = new Promise(function(resolve, reject) {
            reject();
        });
        var getDataUri = jasmine.createSpy('getDataUri').and.returnValue(imagePromise);
        mocPlugin.__set__('getDataUri', getDataUri);

        var contextToServer = jasmine.createSpy('contextToServer').and.returnValue('/foobar');
        mocPlugin.__set__('contextToServer', contextToServer);
        var parseXmlText = jasmine.createSpy('parseXmlText').and.returnValue('/baz.jpg');
        mocPlugin.__set__('parseXmlText', parseXmlText);

        var fetchContent = mocPlugin.__get__('fetchContent');
        fetchContent('/foo', 'bar', widgetModel.id)
            .catch(function() {
                expect(contextToServer).toHaveBeenCalledWith('/foo');
                expect(parseXmlText).toHaveBeenCalledWith('/foo/content/bbp/repositories/image.jpg');
                expect(getDataUri).toHaveBeenCalledWith('/foobar/baz.jpg');
                done();
            });
    });

    it('finds CS references (images) and tries to resolve them - ok', function(done) {
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve({
                status: '200',
                text: function() {
                    return 'content template - this is an image reference: "/foo/content/bbp/repositories/image.jpg"'
                }
            });
        });
        var fetch = jasmine.createSpy('fetch').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetch', fetch);

        var imagePromise = new Promise(function(resolve, reject) {
            resolve('base64 encoded image');
        });
        var getDataUri = jasmine.createSpy('getDataUri').and.returnValue(imagePromise);
        mocPlugin.__set__('getDataUri', getDataUri);

        var contextToServer = jasmine.createSpy('contextToServer').and.returnValue('/foobar');
        mocPlugin.__set__('contextToServer', contextToServer);
        var parseXmlText = jasmine.createSpy('parseXmlText').and.returnValue('/baz.jpg');
        mocPlugin.__set__('parseXmlText', parseXmlText);

        var fetchContent = mocPlugin.__get__('fetchContent');
        fetchContent('/foo', 'bar', widgetModel.id)
            .then(function(content) {
                expect(content).toEqual('content template - this is an image reference: "base64 encoded image"');
                done();
            });
    });
});

describe('mobile optimised content plugin - resolveContent', function () {
    it('reads from the cache - fail', function(done) {
        var readPromise = new Promise(function(resolve, reject) {
            reject('broken');
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .catch(function() {
                done();
            });
    });

    it('checks the cachekey and returns the content if it matches', function(done) {
        var mockContent = {
            cacheKey: '12345',
            data: 'mock content'
        };
        var readPromise = new Promise(function(resolve, reject) {
            resolve(mockContent);
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .then(function(content) {
                expect(content).toEqual(mockContent.data);
                done();
            });
    });

    it('checks the cachekey and fetches fresh content if it doesn\'t match', function(done) {
        var mockContent = {
            cacheKey: '12346',
            data: 'mock content'
        };
        var readPromise = new Promise(function(resolve, reject) {
            resolve(mockContent);
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve('fetchContent');
        });
        var fetchContent = jasmine.createSpy('fetchContent').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetchContent', fetchContent);

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .then(function(content) {
                expect(fetchContent).toHaveBeenCalledWith('/foo', 'bar', widgetModel.id);
                expect(content).toEqual('fetchContent');
                done();
            });
    });

    it('saves the fresh content in the cache', function(done) {
        var mockContent = {
            cacheKey: '12346',
            data: 'mock content'
        };
        var readPromise = new Promise(function(resolve, reject) {
            resolve(mockContent);
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve('fetchContent');
        });
        var fetchContent = jasmine.createSpy('fetchContent').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetchContent', fetchContent);
        var writeStorage = jasmine.createSpy('writeStorage');
        mocPlugin.__set__('writeStorage', writeStorage);

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .then(function(content) {
                expect(writeStorage).toHaveBeenCalledWith(widgetModel.id, {
                    cacheKey: '12345',
                    data: content
                });
                done();
            });
    });

    it('doesn\'t save the fresh content if it fails to fetch it', function(done) {
        var mockContent = {
            cacheKey: '12346',
            data: 'mock content'
        };
        var readPromise = new Promise(function(resolve, reject) {
            resolve(mockContent);
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);
        var fetchPromise = new Promise(function(resolve, reject) {
            reject();
        });
        var fetchContent = jasmine.createSpy('fetchContent').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetchContent', fetchContent);

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .catch(function() {
                done();
            });
    });

    it('fetches fresh content if \'contentsCached\' property is set to false', function(done) {
        var mockContent = {
            cacheKey: '12346',
            data: 'mock content'
        };
        var readPromise = new Promise(function(resolve, reject) {
            resolve(mockContent);
        });
        var readStorage = jasmine.createSpy('readStorage').and.returnValue(readPromise);
        mocPlugin.__set__('readStorage', readStorage);
        var fetchPromise = new Promise(function(resolve, reject) {
            resolve('fetchContent');
        });
        var fetchContent = jasmine.createSpy('fetchContent').and.returnValue(fetchPromise);
        mocPlugin.__set__('fetchContent', fetchContent);
        var deleteFromStorage = jasmine.createSpy('deleteFromStorage');
        mocPlugin.__set__('deleteFromStorage', deleteFromStorage);

        widgetModel.preferences.contentsCached = {
            name: 'contentsCached',
            value: 'false'
        };

        var resolveContent = mocPlugin.__get__('resolveContent');
        resolveContent('/foo', 'bar', widgetModel)
            .then(function(content) {
                expect(fetchContent).toHaveBeenCalledWith('/foo', 'bar', widgetModel.id);
                expect(deleteFromStorage).toHaveBeenCalledWith(widgetModel.id);
                expect(content).toEqual('fetchContent');
                done();
            });
    });
});

describe('mobile optimised content plugin - postRender', function () {
    it('searches for g:includes', function() {
        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([]);

        moc.postRender(widgetInstance, null, widgetModel);

        expect(widgetInstance.body.getElementsByTagName).toHaveBeenCalledWith('g:include');
    });

    var otherInclude = {};
    otherInclude.getAttribute = jasmine.createSpy('getAttribute').and.returnValue('/somewhere');

    it('ignores g:includes for other things than ICE / structured content', function() {
        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([otherInclude]);

        moc.postRender(widgetInstance, null, widgetModel);

        expect(otherInclude.getAttribute).toHaveBeenCalledWith('src');
    });

    var cxp55include = {};
    cxp55include.getAttribute = jasmine.createSpy('getAttribute').and.returnValue('/contenttemplates\/rendered');
    cxp55include.querySelectorAll = jasmine.createSpy('querySelectorAll').and.returnValue([1]);

    it('ignores CXP 5.5 g:includes', function() {
        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp55include]);

        moc.postRender(widgetInstance, null, widgetModel);

        expect(cxp55include.querySelectorAll).toHaveBeenCalledWith('[name="templateUrl"]');
    });

    var cxp56include = {
        outerHTML: 'cxp56include with g:include'
    };
    cxp56include.getAttribute = jasmine.createSpy('getAttribute').and.returnValue('/contenttemplates\/rendered');
    cxp56include.querySelectorAll = jasmine.createSpy('querySelectorAll').and.returnValue([]);

    widgetInstance.features.cxp.config.get = jasmine.createSpy('get').and.returnValue('bar');

    it('renames CXP 5.6 g:includes', function() {
        var replaceInclude = jasmine.createSpy('replaceInclude');
        mocPlugin.__set__('replaceInclude', replaceInclude);

        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp56include]);

        moc.postRender(widgetInstance, null, widgetModel);

        expect(replaceInclude).toHaveBeenCalledWith(cxp56include, 'cxp56include with moc:include');
    });

    it('tries to resolve content', function() {
        var contentPromise = new Promise(function(resolve, reject) {
            reject('fail');
        });
        var resolveContent = jasmine.createSpy('resolveContent').and.returnValue(contentPromise);
        mocPlugin.__set__('resolveContent', resolveContent);

        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp56include]);

        moc.postRender(widgetInstance, null, widgetModel);

        expect(resolveContent).toHaveBeenCalledWith('/foo', 'bar', widgetModel);
    });

    it('replaces the g:include if it resolves the content', function(done) {
        var contentPromise = new Promise(function(resolve, reject) {
            resolve('new content');
        });
        var resolveContent = jasmine.createSpy('resolveContent').and.returnValue(contentPromise);
        mocPlugin.__set__('resolveContent', resolveContent);

        var replaceInclude = jasmine.createSpy('replaceInclude').and.returnValue('new node');
        mocPlugin.__set__('replaceInclude', replaceInclude);

        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp56include]);

        moc.postRender(widgetInstance, null, widgetModel);

        setTimeout(function(){
            expect(replaceInclude).toHaveBeenCalledWith('new node', 'new content');
            done();
        }, 100);
    });

    it('sends the \'contentLoaded\' event if it resolves the content', function(done) {
        var contentPromise = new Promise(function(resolve, reject) {
            resolve('ok');
        });
        var resolveContent = jasmine.createSpy('resolveContent').and.returnValue(contentPromise);
        mocPlugin.__set__('resolveContent', resolveContent);

        var sendEvent = jasmine.createSpy('sendEvent');
        mocPlugin.__set__('sendEvent', sendEvent);

        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp56include]);

        moc.postRender(widgetInstance, null, widgetModel);

        setTimeout(function(){
            expect(sendEvent).toHaveBeenCalledWith(widgetInstance.body, 'contentLoaded');
            done();
        }, 100);
    });

    it('sends the \'contentError\' event if it fails to resolve', function(done) {
        var contentPromise = new Promise(function(resolve, reject) {
            reject('fail');
        });
        var resolveContent = jasmine.createSpy('resolveContent').and.returnValue(contentPromise);
        mocPlugin.__set__('resolveContent', resolveContent);

        var sendEvent = jasmine.createSpy('sendEvent');
        mocPlugin.__set__('sendEvent', sendEvent);

        var moc = mocPlugin('/foo');
        widgetInstance.body.getElementsByTagName = jasmine.createSpy('getElementsByTagName').and.returnValue([cxp56include]);

        moc.postRender(widgetInstance, null, widgetModel);

        setTimeout(function(){
            expect(sendEvent).toHaveBeenCalledWith(widgetInstance.body, 'contentError');
            done();
        }, 100);
    });
});