/* globals Promise: false */
'use strict';

// Jasmine mock f*ckery
var window = global.window; // jshint ignore:line

/*
Mobile Optimised Content (moc)

Description

This plugin caches ICE and structured content (5.6 content only!).

Usage

Include this plugin before the BackbaseFormatPlugin. If you include it
after, it doesn't do anything because the BackbaseFormatPlugin has
replaced the content g:includes with placeholders.

Internals

This plugin hooks into the postRender phase (it needs to have the widget
start file). It finds all ICE and structured content g:includes, and
renames them to moc:include so the BackbaseFormatPlugin cannot find them
anymore.

After this it checks the device cache (PersistentStorage) or localStorage
for the cached ICE / structured content template. If it finds a cached
template, it checks for expired caches using the widgetContentsUpdated
property.

If the cache is empty or expired, the plugin fetches the ICE /
structured content template itself. After this it finds all CS content
urls and fetches the contents of those urls and transforms them into
data-uris. This allows us to cache just the template. After that it
saves the template to PersistentStorage or localStorage. This plugin caches
successful responses only!

Widget preferences

There are some optional widget preferences that influence the caching.
 - widgetContentsUpdated    added and updated by ICE widgets
 - contentsCached           set this to false to disable caching

Events and activity indicators

This plugin doesn't replace the g:include with a placeholder, it just
renames it. So you can add a class or nest some image or other content
inside the g:include to show an activity indicator. After the content
is resolved, it replaces the g:include so the indicator will dissapear
automatically.

It is also possible to do this programatically. In your g:onload handler
you can display a spinner, and listen for the 'contentLoaded' and
'contentError' events by binding handlers to the widgetInstance.body.
In the handlers you can remove the spinner and show the content, or show
an error if the content cannot be fetched.
 */

var fetch = require('cxp-web-apis/src/util/deduping-fetch');

/*
 *  helpers
 */
var mimeTypes = {
    png: 'image/png',
    jpg: 'image/jpeg',
    jpeg: 'image/jpeg',
    gif: 'image/gif',
    html: 'text/html',
    js: 'application/javascript',
    css: 'text/css'
};

var toArray = function(htmlCollection) {
    return [].slice.call(htmlCollection);
};

var htmlToElement = function(html) {
    var template = window.document.createElement('template');
    template.innerHTML = html;
    return template.content.firstChild;
};

var replaceInclude = function(include, html) {
    var parent = include.parentNode;
    var newInclude = htmlToElement(html);

    parent.replaceChild(newInclude, include);

    return newInclude;
};

var createContentUrl = function(contextRoot, portalName, id) {
    return contextRoot + '/contenttemplates/rendered?' +
        'contextItemName=' + portalName + '&' +
        'uuid=' + id;
};

var parseXmlText = function(text) {
    var parsed = (new window.DOMParser())
        .parseFromString('<foo>' + text + '</foo>', 'text/xml');
    return parsed && parsed.childNodes &&
        parsed.childNodes[0] && parsed.childNodes[0].textContent;
};

var urlToMimeType = function(url) {
    var extension = url.match('\.([^.]+)$')[1];
    return mimeTypes[extension];
};

var sendEvent = function(domNode, eventName) {
    var event = window.document.createEvent('HTMLEvents');
    event.initEvent(eventName, true, true);
    domNode.dispatchEvent(event);
};

var contextToServer = function(contextRoot) {
    return contextRoot.match(/^[^/]+\/\/[^/]+/) || '';
};

/*
 *  storage functions
 */
var mangleName = function(id) {
    return 'moc-cache-' + id;
};

var readStorage = function(id, cb) {
    var storage = window.localStorage;
    var persistentStorage = window.cxp && window.cxp.mobile &&
        window.cxp.mobile.plugins && window.cxp.mobile.plugins.PersistentStorage;

    return new Promise(function(resolve, reject) {
        if (persistentStorage) {
            persistentStorage.getItem(
                function(data){
                    if(data){
                        resolve(JSON.parse(data));
                    }else{
                        resolve();
                    }
                },
                // we do not reject the promise if the id isn't found in
                // storage, we just return undefined
                function() {
                    resolve();
                },
                mangleName(id)
            );
        } else {
            var data = storage[mangleName(id)];
            if (data) {
                resolve(JSON.parse(data));
            } else {
                // we do not reject the promise if the id isn't found in
                // storage, we just return undefined
                resolve();
            }
        }
    });
};

var writeStorage = function(id, content) {
    var storage = window.localStorage;
    var persistentStorage = window.cxp && window.cxp.mobile &&
        window.cxp.mobile.plugins && window.cxp.mobile.plugins.PersistentStorage;

    return new Promise(function(resolve, reject) {
        if (persistentStorage) {
            persistentStorage.setItem(
                resolve,
                reject,
                mangleName(id),
                JSON.stringify(content)
            );
        } else {
            storage[mangleName(id)] = JSON.stringify(content);
            resolve();
        }
    });
};

var deleteFromStorage = function(id) {
    var storage = window.localStorage;
    var persistentStorage = window.cxp && window.cxp.mobile &&
        window.cxp.mobile.plugins && window.cxp.mobile.plugins.PersistentStorage;

    return new Promise(function(resolve, reject) {
        if (persistentStorage) {
            persistentStorage.removeItem(
                resolve,
                reject,
                mangleName(id)
            );
        } else {
            storage.removeItem(mangleName(id));
            resolve();
        }
    });
};


/*
 *  Promise based functions
 */

// taken from https://davidwalsh.name/convert-image-data-uri-javascript
// changed to work with promises and to use proper mimeTypes
var getDataUri = function(url) {
    return new Promise(function(resolve, reject) {
        var image = new window.Image();

        image.onload = function () {
            var canvas = window.document.createElement('canvas');
            canvas.width = this.naturalWidth;
            canvas.height = this.naturalHeight;

            canvas.getContext('2d').drawImage(this, 0, 0);

            resolve(canvas.toDataURL(urlToMimeType(url)));
        };

        image.onerror = function () {
            reject();
        };

        image.src = url;
    });
};

var fetchContent = function(contextRoot, portalName, id) {
    return fetch(createContentUrl(contextRoot, portalName, id))
        .then(function(res) {
            return res.status >= 200 && res.status < 300 ?
                Promise.resolve(res.text()) : Promise.reject(res.statusText);
        })
        .then(function(html) {
            // find all CS content urls to resolve
            return html
                .split(/([^"]+\/content\/atom\/[^"]+)|([^"]+\/content\/bbp\/repositories\/[^"]+)/g);
        })
        .then(function(chunks) {
            return Promise.all(
                chunks
                    .map(function(chunk) {
                        if (/\/content\/atom\/|\/content\/bbp\/repositories\//.test(chunk)) {
                            return getDataUri(
                                contextToServer(contextRoot) +
                                // CS content urls seem to be XML encoded. Fix that here
                                parseXmlText(chunk)
                            );
                        }
                        return chunk;
                    })
            );
        })
        .then(function(chunks) {
            return chunks
                .join('');
        });
};

var resolveContent = function(contextRoot, portalName, model) {
    return readStorage(model.id)
        .then(function(content) {
            var cacheKey = model.preferences.widgetContentsUpdated ? model.preferences.widgetContentsUpdated.value : 0;
            var contentsCached = model.preferences.contentsCached ? model.preferences.contentsCached.value === 'true' : true;

            // remove the previously cached element if the cache has been disabled for it.
            if (contentsCached === false){
                deleteFromStorage(model.id);
            }

            // ICE widget - proper expiry detection using 'widgetContentsUpdated' preference in the model
            if (contentsCached && content && content.cacheKey === cacheKey) {
                return content.data;

            // fetch content ourselves
            } else {
                return fetchContent(contextRoot, portalName, model.id)
                    .then(function(content) {
                        if(contentsCached){
                            writeStorage(model.id, {
                                cacheKey: cacheKey,
                                data: content
                            });
                        }
                        return content;
                    });
            }
        });
};

module.exports = function(contextRoot) {
    return {
        name: 'mobile-optimized-content-plugin',
        postRender: function(widgetInstance, widgetRenderer, widgetModel) {
            var portalName = widgetInstance.features.cxp.config.get('portalName');

            var includes = toArray(widgetInstance.body.getElementsByTagName('g:include'))
                .filter(function(include) {
                    return /\/contenttemplates\/rendered/.test(include.getAttribute('src')) &&
                        include.querySelectorAll('[name="templateUrl"]').length === 0;
                })
                .map(function(include) {
                    // replace the g:include with moc:include so the BackbaseFormatPlugin doesn't resolve it
                    var domNode = replaceInclude(include, include.outerHTML.replace(/g:include/g, 'moc:include'));

                    return resolveContent(contextRoot, portalName, widgetModel)
                        .then(function(content) {
                            replaceInclude(domNode, content);
                        });
                });

            if (includes.length) {
                Promise.all(includes)
                    // these fire only after all content g:includes have finished
                    .then(function() {
                        sendEvent(widgetInstance.body, 'contentLoaded');
                    })
                    .catch(function() {
                        sendEvent(widgetInstance.body, 'contentError');
                    });
            }

            return widgetInstance;
        }
    };
};
